"""
    Assetic.localdb_sync.py
    Tools to facilitate bulk export of data from assetic and
    syncing exported data to a local sql server database
    
    Checks if there are files to download from from Export All task
    Downloads the file and saves locally to then process
    Uses file system to record pending & processed export jobs
    Data is written to a temp db table and then the 'merge' SQL is used
    to update/insert records in the permanant table
    Will create the tables as required (TODO check columns exist)
    Writes to a log table the changed record (TODO expand the output)
"""
#from __future__ import absolute_import

import pypyodbc
import sys
import assetic   #This is an SDK to wrap the REST endpoints
import datetime
import dateutil.parser
import logging
import os
import shutil
from contextlib import contextmanager
from assetic.tools import BulkProcesses
from pprint import pformat
import six
import time
try:
    import configparser as configparser
except:
    #V2 python?
    import ConfigParser as configparser
import csv
from assetic.api_client import ApiClient
from assetic.configuration import Configuration

class SyncToLocalProcesses(object):
    """
    Class to manage processes that sync Assetic search data to a local DB
    """

    def __init__(self,configfile = None,workingdir=None,debug=False):
        if configfile == None:
            configfile = os.path.join(os.getcwd(), 'assetic.ini')
        
        if os.path.isfile(configfile) == False:
            message = "Configuration file not found.  Expecting assetic.ini in current folder, or set via module init"
            print(message)
            raise ValueError(message)
        else:
            self.configfile = configfile

        if workingdir == None:
            workingdir = os.getcwd()
        if os.path.isdir(workingdir) == False:
            message = "Working directory [{0}] does not exist".format(workingdir)
            raise ValueError(message)
        
        # Read ini
        settings = configparser.ConfigParser()
        configfile = self.configfile
        settings.read(configfile)
        # Apply config settings
        conf = Configuration()
        conf.host = settings.get('environment', 'url')
        conf.username = settings.get('auth', 'username')
        conf.password = settings.get('auth', 'api_key')
        if debug == True:
            conf.logger_file =  + + '//LocalSyncProcessesLog.txt'
            logging.basicConfig() 
            logging.getLogger().setLevel(logging.DEBUG)
        self.conf = conf
        self.auth = conf.get_basic_auth_token()
        self.workingdir = workingdir       

    def create_export_task(searchid):
        """
        Given a search profile initiate a bulk export job
        Returned: The GUID of the job (task)
        """
        ##create an instance of the api client
        client = assetic.ApiClient(_self.conf.host,"Authorization",self.auth)

        ##api not yet implemented
        return None
    

    def get_export_docid(exporttaskguid,savefolder):
        """
        Given a export job GUID check status
        Returned: A tuple (2 fields)
        1. The status or None if profile not found
        2. docid if ready for download, otherwise None
        """
        ##create an instance of the api client
        client = assetic.ApiClient(_self.conf.host,"Authorization",self.auth)

        ##api not yet implemented
        return None,None

    def get_document(docid,savefolder):
        """
        Returned: the csv file path(full path) for a given document id
        """
        
        ##download the file
        bulkprocesses = BulkProcesses(self.configfile)
        csvfile = bulkprocesses.document_get(savefolder,docid)
        ##return the file name
        return csvfile

    def get_column_metadata(self):
        """
        For a given search profile get the column metadata
        Returned: a dictionary tuple with internal name, field label, datatype
        """
        ##create an instance of the api client
        client = assetic.ApiClient(_self.conf.host,"Authorization",self.auth)

        return None

    def initiate_export_process(self,searchid):
        """
        For a given search profile initiate the bulk export
        Returned: True = Success, False = Error
        Also create a file named after the tak id.  This file is in the
        'inprogress' folder.
        Later we can check this folder for 'inprogress' tasks
        """
        taskid = create_export_task(searchid)
        if taskid == None:
            return False
        ##create an empty file in the working 'in progress' folder
        ##filename is the taskid
        file = self.workingdir + "/inprogress/" + taskid 
        try:
            open(file,'w').close()
        except:
            print("Unable to save file to {0}".format(file))
            return False
        return True

    def get_csv_if_ready_process(self):
        """
        Look in the working folder 'inprogress' for file
        The file name is the guid of a job.  Check if job is complete
        and if so, download the file to the 'awaitingsync' folder
        """
        sourcedir = self.workingdir + "/inprogress"
        downloaddir = self.workingdir + "/csvexports"
        processedjobs = self.workingdir + "/processedjobs"
        files = os.listdir(sourcedir)
        for filename in files:
            csvfile = getdocument(filename,downloaddir)
            if csvfile != None:
                shutil.move(sourcedir + "/" + filename,processedjobs + "/" + filename)
            else:
                shutil.move(sourcedir + "/" + filename,downloaderror + "/" + filename)

    def dbsync_csvfile_process(self,dbserver,dbname,profiles):
        """
        Look in the working folder 'csvexports' for file
        The file name is the profile name of the job.
        Upload to database and move to csvsynced folder
        """
        sourcedir = self.workingdir + "/inprogress/"
        downloaddir = self.workingdir + "/csvexports"
        processedcsv = self.workingdir + "/processedcsv"
        errorcsv = self.workingdir + "/errorcsv"


        files = os.listdir(downloaddir)
        if len(files) == 0:
            return
        dbtools = DB_Tools(dbserver,dbname)
        for filename in files:
            print(filename)
            fullfilename = downloaddir + "/" + filename
            for profile in profiles:
                if profile.profilename + ".csv" == filename:
                    print("profile: {0},file: {1}".format(profile.profilename,filename))
                    dbtable = profile.tablename
                    keyfield = profile.keyfield
                    logtable = profile.logtable
                    success = dbtools.csv_to_db(fullfilename,dbtable,keyfield,logtable)
                    #dt = datetime.datetime.utcnow().isoformat()
                    dt = str(int(time.time()))
                    filename = dt + "_" + filename
                    if success:
                        shutil.move(fullfilename,processedcsv + "/" + filename)
                    else:
                        shutil.move(fullfilename,errorcsv + "/" + filename)

class SearchProfileRepresentation(object):
    """"
    A structure for defining table metadata and relationships between
    search profile names, id's and tables
    """
    def __init__(self,profileguid=None,profilename=None,tablename=None,keyfield=None,logtable=None):
        """
        SearchProfileRepresentation - a model defining metadata for syncing a
        search profile to a local database table via bulk export api
        """
        self.fieldtypes = {
            "profileguid": "str",
            "profilename": "str",
            "tablename": "str",
            "keyfield": "str",
            "logtable": "str",
        }
        self._profileguid = profileguid
        self._profilename = profilename
        self._tablename = tablename
        self._keyfield = keyfield
        self._logtable = logtable
        
    @property
    def profileguid(self):
        return self._profileguid
    @profileguid.setter
    def profileguid(self,profileguid):
        self._profileguid = profileguid
    
    @property
    def profilename(self):
        return self._profilename
    @profilename.setter
    def profilename(self,profilename):
        self._profilename = profilename

    @property
    def tablename(self):
        return self._tablename
    @tablename.setter
    def tablename(self,tablename):
        self._tablename = tablename

    @property
    def keyfield(self):
        return self._keyfield
    @keyfield.setter
    def keyfield(self,keyfield):
        self._keyfield = keyfield

    @property
    def logtable(self):
        return self._logtable
    @logtable.setter
    def logtable(self,logtable):
        self._logtable = logtable
    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in six.iteritems(self.fieldtypes):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

class SearchProfileColumnRepresentation(object):
    """"
    A structure for defining table metadata and relationships between
    search profile names, id's and tables
    """
    def __init__(self,profileguid=None,profilename=None,tablename=None,keyfield=None,logtable=None):
        """
        SearchProfileRepresentation - a model defining metadata for syncing a
        search profile to a local database table via bulk export api
        """
        self.fieldtypes = {
            "internalname": "str",
            "label": "str",
            "datatype": "str",
            "fieldlength": "str"
        }
        self._internalname = internalname
        self._label = label
        self._datatype = datatype
        self._fieldlength = fieldlength
        
    @property
    def internalname(self):
        return self._internalname
    @internalname.setter
    def internalname(self,internalname):
        self._internalname = internalname
    
    @property
    def label(self):
        return self._label
    @label.setter
    def label(self,label):
        self._label = label

    @property
    def datatype(self):
        return self._datatype
    @datatype.setter
    def datatype(self,datatype):
        self._datatype = datatype

    @property
    def fieldlength(self):
        return self._fieldlength
    @fieldlength.setter
    def fieldlength(self,fieldlength):
        self._fieldlength = fieldlength

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in six.iteritems(self.fieldtypes):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other

class DB_Tools(object):
    """
    Class to manage connections to databases and typical processes
    """
    def __init__(self,dbserver,dbname):
        
        self.connstr = "Driver={SQL Server};" + """Server={0};Database={1};
            Trusted_Connection=True;
            APP=Assetic Asset Sync""".format(dbserver,dbname)                 

    @contextmanager
    def open_db_connection(self, commit=False):
        """
        Use contextmanager to manage connections so we cleanup connection
        as we go to avoid database locking issues on exception
        """
        connection = pypyodbc.connect(self.connstr)
        cursor = connection.cursor()
        try:
            yield cursor
        except pypyodbc.DatabaseError as err:
            cursor.execute("ROLLBACK")
            raise err
        except Exception as ex:
            print(ex)
        else:
            if commit:
                cursor.execute("COMMIT")
            else:
                cursor.execute("ROLLBACK")
        finally:
            connection.close()

        
    def csv_to_db(self,csvfile,dbtable,keyfield,logtable):
        #write data in file to database
        #get list of columns
        bIsHeader = True
        allrows = []
        saved_row = []

        with open(csvfile, 'rt', encoding='utf-8', newline = '') as csvfile:
            readCSV = csv.reader(csvfile,delimiter=',')
            for row in readCSV:
                if bIsHeader == True:
                    bIsHeader = False
                    columns = row
                else:
                    #cater for carriage returns in a column
                    if len(saved_row):
                        row_tail = saved_row.pop()  #gets last item in the list
                        row[0] = row_tail + '\r' +row[0]  # reconstitute field broken by newline
                        row = saved_row + row       # and reassemble the row (possibly only partially)
                    if len(row) >= len(columns):
                        allrows.append(row)
                        saved_row = []
                    else:
                        saved_row = row
                        
        #name for tmp table we'll write data dump to
        tmpdbtable = "#{0}_tmp".format(dbtable)

        ##First get the database table column names and build the various tables
        ##TODO add a check for existing tables to make sure all columns present
        insertcolumns = []
        updatecolumns = []
        parammark = []  #add a "?" for each column to use in sql parameter settings
        for column in columns:
            insertcolumns.append("[{0}]".format(column))
            updatecolumns.append("a.[{0}]=c.[{0}]".format(column))
            parammark.append("?")

        sqltblchktmp =  "select object_id('{0}')".format(tmpdbtable)
        sqltruncatetmp = "truncate table {0}".format(tmpdbtable)
        sqlcreatetmp = """create table {0} ({1} nvarchar(500) COLLATE
            database_default)""".format(tmpdbtable,
            " nvarchar(500) COLLATE database_default,".join(insertcolumns ))
        sqltblchk =  "select object_id('{0}')".format(dbtable)
        sqlcreate = """create table {0} ({1} nvarchar(500))
            """.format(dbtable,"nvarchar(500),".join(insertcolumns ))
        sqllogchk = "select object_id('{0}')".format(logtable)
        sqllogcreate = """CREATE TABLE {0}(
            [action] [nvarchar](500),
            {1} [nvarchar](500),
            [syncdate] [datetime])""".format(logtable,keyfield)
        ##Now create insert string
        sqlinsert = "insert into {0} (".format(tmpdbtable)
        sqlinsert = sqlinsert + ",".join(insertcolumns ) + ") values ("
        sqlinsert = sqlinsert + ",".join(parammark ) + ")"

        ##Now add data to sql insert string and do insert
        ##(using 'insertmany' effectively row by row)

        ##now refresh base table from temp table where there are changes
        ##write the change to the archive file
        ##TODO output all columns for updated record (original record)
        sqldiff = """
        WITH diff as ( select {2}
            FROM {0}
            except select {2}
            FROM {1}
            )
            MERGE {1} as a using diff as c on
            (c.{5}=a.{5})
            when MATCHED then update set {3}
            When not matched then insert
            ({2}) 
            values ({2})
            output $action,inserted.{5},getdate() into {4};
        """.format(tmpdbtable,dbtable,",".join(insertcolumns),
                   ",".join(updatecolumns),logtable,keyfield)

        ##Now execute all the sql statements in the one transaction block
        ##Bit slow opening & closing db connection, but transaction block works well
        try:
            with self.open_db_connection(True) as cursor:
                #check if temportaty table exists
                print(sqltblchktmp)
                cursor.execute(sqltblchktmp)
                row = cursor.fetchone()
                if row[0] == None:
                    print(sqlcreatetmp)
                    cursor.execute(sqlcreatetmp)
                else:
                    print(sqltruncatetmp)
                    cursor.execute(sqltruncatetmp)

                #check if permanent table exists
                cursor.execute(sqltblchk)
                print(sqltblchk)
                row = cursor.fetchone()
                if row[0] == None:
                    #table not found - create permament table for current data
                    print(sqlcreate)
                    cursor.execute(sqlcreate)
                #check if log table exists
                print(sqllogchk)
                cursor.execute(sqllogchk)
                row = cursor.fetchone()
                if row[0] == None:
                    #table not found - create permament table for logging changes
                    print(sqllogcreate)
                    cursor.execute(sqllogcreate)
                ##write all the data to the temp table in db.
                print(sqlinsert)
                cursor.executemany(sqlinsert,allrows)
                ##execute diff statement to populate permanent tables from temp table
                print(sqldiff)
                cursor.execute(sqldiff)
        except Exception as ex:
            print(ex)
            return False
        return True
